from __future__ import annotations

import io
import json
import tarfile
import uuid
from abc import ABCMeta, abstractmethod
from datetime import datetime, timezone
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Any, BinaryIO, cast
from typing_extensions import Self

import requests
from requests.auth import HTTPBasicAuth

from eportal.patchsource import PatchSource
from eportal.utils import get_proxy_settings, get_instance_id


class DataPackage(metaclass=ABCMeta):
    MANIFEST_SCHEMA_VERSION: int = 1

    def __init__(self) -> None:
        self._tar: tarfile.TarFile | None = None

    @property
    @abstractmethod
    def data_type(self) -> str: ...  # pragma: no cover

    @property
    @abstractmethod
    def upload_uri(self) -> str: ...  # pragma: no cover

    def __enter__(self) -> Self:
        for compression_mode in ('w:xz', 'w:bz2', 'w:gz'):
            tmpfile = NamedTemporaryFile(suffix=f'.tar.{compression_mode[2:]}', delete=False)
            tmpfile.close()
            try:
                self._tar = tarfile.open(name=tmpfile.name, mode=compression_mode)
                self._add_manifest()
                return self
            except Exception as err:
                Path(tmpfile.name).unlink(missing_ok=True)
                self._tar = None
                if not isinstance(err, tarfile.CompressionError):
                    raise
        raise tarfile.CompressionError('No supported compression method found')

    def __exit__(self, exc_type: type[BaseException] | None, exc_value: BaseException | None, traceback: Any) -> None:
        cast(tarfile.TarFile, self._tar).close()
        if exc_type is not None:
            self.remove()
            self._tar = None

    def add_file(
        self, path: str, fileobj: BinaryIO | None = None, filepath: Path | None = None, data: bytes | None = None
    ) -> None:
        tar = self._ensure_tar_created()
        path = self._normalize_path(path)
        if filepath is not None:
            if not filepath.exists():
                raise FileNotFoundError(f'File not found: {filepath}')
            tar.add(filepath, arcname=path)
        elif fileobj is not None:
            info = tarfile.TarInfo(path)
            info.size = fileobj.seek(0, io.SEEK_END)
            fileobj.seek(0)
            tar.addfile(info, fileobj)
        elif data is not None:
            info = tarfile.TarInfo(path)
            info.size = len(data)
            tar.addfile(info, io.BytesIO(data))
        else:
            raise ValueError('No fileobj, filepath or data provided')

    def add_json(self, path: str, obj: dict[str, Any], compact: bool = False) -> None:
        def encoder_stub(obj: Any) -> str:
            return f'Failed to encode object: {obj!r}'

        self.add_file(
            path, data=json.dumps(obj, ensure_ascii=False, indent=None if compact else 4, default=encoder_stub).encode()
        )

    def _add_manifest(self) -> None:
        manifest = {
            'schema_version': self.MANIFEST_SCHEMA_VERSION,
            'type': self.data_type,
            'instance_id': get_instance_id(),
            'time_created': datetime.now(tz=timezone.utc).isoformat(),
        }
        self.add_json('manifest.json', manifest)

    def _normalize_path(self, path: str) -> str:
        return str(Path('/' + path).resolve().relative_to('/'))

    def _ensure_tar_created(self) -> tarfile.TarFile:
        if not self._tar:
            raise RuntimeError('DataPackage should be used as a context manager')
        return self._tar

    @property
    def path(self) -> Path:
        tar = self._ensure_tar_created()
        return Path(cast(str, tar.name))

    def send(self) -> str:
        tar = self._ensure_tar_created()
        tar.close()

        ps = PatchSource()
        ps.check_configuration()

        path = self.path
        package_name = str(uuid.uuid4()) + path.name[path.name.find('.') :]
        with open(path, 'rb') as fileobj:
            response = requests.put(
                ps.get_url(f'{self.upload_uri}/{package_name}'),
                auth=HTTPBasicAuth(*ps.auth),
                headers={'content-type': 'application/octet-stream'},
                data=fileobj,
                timeout=30,
                proxies=get_proxy_settings(),
            )
            response.raise_for_status()
            return package_name

    def remove(self) -> None:
        if self._tar:
            self.path.unlink(missing_ok=True)


class DoctorPackage(DataPackage):
    data_type = 'eportal-doctor'
    upload_uri = 'upload/eportal-doctor'


class CheckinPackage(DataPackage):
    data_type = 'eportal-checkins'
    upload_uri = 'upload/eportal-checkins'
